/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *  MA 02110-1301, USA.
 *
 *  Copyright © 2006 Pierre Habouzit
 */

/*
 * Copyright notice from original mutt:
 * Copyright (C) 1996-2000 Michael R. Elkins <me@mutt.org>
 *
 * This file is part of mutt-ng, see http://www.muttng.org/.
 * It's licensed under the GNU General Public License,
 * please see the file GPL in the top level source directory.
 */

#include <lib-lib/lib-lib.h>

#include "recvattach.h"
#include "charset.h"
#include "mime.h"

/* Reads an arbitrarily long header field, and looks ahead for continuation
 * lines.  ``line'' must point to a dynamically allocated string; it is
 * increased if more space is required to fit the whole line.
 */
ssize_t mutt_read_rfc822_line(FILE *f, char **line, ssize_t *n)
{
    ssize_t pos = 0;

    for (;;) {
        char *p = *line;

        /* end of file or end of headers */
        if (!fgets(p + pos, *n - pos, f) || (ISSPACE(*p) && pos == 0)) {
            *p = '\0';
            return 0;
        }

        pos += m_strlen(p + pos);
        if (p[pos - 1] == '\n') {
            int c;

            /* remove trailing spaces. safe: p[0] is not a space */
            do {
                p[--pos] = '\0';
            } while (ISSPACE(p[pos]));

            /* check to see if the next line is a continuation line */
            c = fgetc(f);
            if (c != ' ' && c != '\t') {
                /* next line is a separate header field or EOH */
                ungetc(c, f);
                return pos;
            }

            /* eat tabs and spaces from the beginning of the continuation line */
            do {
                c = fgetc(f);
            } while (c == ' ' || c == '\t');
            ungetc(c, f);

            /* string is still terminated because we removed at least one
               whitespace char above */
            p[pos++] = ' ';
        }

        if (*n < pos + STRING) {
            /* grow the buffer */
            *n += STRING;
            p_realloc(line, *n);
        }
    }
}

/* TODO: Make that a string list somehow */
string_list_t *mutt_parse_references(char *s, int in_reply_to)
{
    string_list_t *lst = NULL;
    int n = 0;
    char *o = NULL;

    /* some mail clients add other garbage besides message-ids, so do a quick
     * check to make sure this looks like a valid message-id
     * some idiotic clients also break their message-ids between lines, deal
     * with that too (give up if it's more than two lines, though)
     */

    for (s = strtok(s, " \t;"); s; s = strtok(NULL, " \t;")) {
        char *new = NULL;

        if (*s == '<') {
            n = m_strlen(s);
            if (s[n - 1] != '>') {
                o = s;
                continue;
            }

            new = m_strdup(s);
        } else if (o) {
            ssize_t m = m_strlen(s);

            if (s[m - 1] != '>') {
                o = NULL;
            } else {
                new = p_new(char, n + m + 1);
                m_strcpy(new, n + m + 1, o);
                m_strcpy(new + n, m + 1, s);
            }
        }

        /* make sure that this really does look like a message-id.
         * it should have exactly one @, and if we're looking at
         * an in-reply-to header, make sure that the part before
         * the @ has more than eight characters or it's probably
         * an email address
         */
        if (new) {
            char *at = strchr(new, '@');
            string_list_t *tmp;

            if (!at || strchr(at + 1, '@') || (in_reply_to && at - new <= 8)) {
                p_delete(&new);
                continue;
            }

            tmp = p_new(string_list_t, 1);
            tmp->data = new;
            tmp->next = lst;
            lst = tmp;
        }
    }

    return lst;
}

int mutt_check_encoding(const char *s)
{
    int tok = mime_which_token(s, -1);
    switch (tok) {
      case MIME_7BIT:
        return ENC7BIT;
      case MIME_8BIT:
        return ENC8BIT;
      case MIME_BINARY:
        return ENCBINARY;
      case MIME_QUOTED_PRINTABLE:
        return ENCQUOTEDPRINTABLE;
      case MIME_BASE64:
        return ENCBASE64;
      case MIME_X_UUENCODE:
        return ENCUUENCODED;
      default:
        return ENCOTHER;
    }
}

int mutt_check_mime_type(const char *s)
{
    int tok;

    if (!m_strcmp(s, "*") || !m_strcmp(s, ".*"))
        return TYPEANY;

    tok = mime_which_token(s, -1);
    switch (tok) {
      case MIME_TEXT:        return TYPETEXT;
      case MIME_MULTIPART:   return TYPEMULTIPART;
      case MIME_APPLICATION: return TYPEAPPLICATION;
      case MIME_MESSAGE:     return TYPEMESSAGE;
      case MIME_IMAGE:       return TYPEIMAGE;
      case MIME_AUDIO:       return TYPEAUDIO;
      case MIME_VIDEO:       return TYPEVIDEO;
      case MIME_MODEL:       return TYPEMODEL;
      default:               return TYPEOTHER;
    }
}

static parameter_t *parse_parameters(const char *s)
{
    parameter_t *res = NULL;
    parameter_t **list = &res;

    while (*s) {
        const char *p;
        parameter_t *new;
        int i;

        s = skipspaces(s);
        if (*s == '=')             /* parameters are fucked up, go away */
            break;

        p = strpbrk(s, "=;");
        if (!p)
            break;

        if (*p == ';') {
            /* if we hit a ; now the parameter has no value, just skip it */
            s = p + 1;
            continue;
        }

        i = p - s;
        new = parameter_new();
        new->attribute = p_dupstr(s, i);

        while (--i >= 0 && ISSPACE(new->attribute[i])) {
            new->attribute[i] = '\0';
        }
        s = skipspaces(p + 1);                      /* skip over the = */

        if (*s == '"') {
            char buffer[LONG_STRING];
            int state_ascii = 1;

            s++;
            for (i = 0; *s && i < ssizeof(buffer) - 1; i++, s++) {
                /* As iso-2022-* has a characer of '"' with non-ascii state,
                 * ignore it. */
                if (*s == 0x1b && i < ssizeof(buffer) - 2) {
                    state_ascii = s[1] == '(' && (s[2] == 'B' || s[2] == 'J');
                }
                if (state_ascii && *s == '"')
                    break;

                if (*s == '\\') {
                    buffer[i] = *++s;
                } else {
                    buffer[i] = *s;
                }
            }

            new->value = p_dupstr(buffer, i);
        } else {
            for (p = s; *p && *p != ' ' && *p != ';'; p++);
            new->value = p_dupstr(s, p - s);
        }

        *list = new;
        list = &new->next;

        s = strchr(s, ';');           /* Find the next parameter */
        if (!s)
            break;                    /* no more parameters */
    }

    rfc2231_decode_parameters(&res);
    return res;
}

void mutt_parse_content_type(char *s, BODY *ct)
{
    char *pc;
    char *subtype;

    p_delete(&ct->subtype);
    parameter_list_wipe(&ct->parameter);

    /* First extract any existing parameters */
    if ((pc = strchr(s, ';')) != NULL) {
        *pc++ = '\0';
        ct->parameter = parse_parameters(vskipspaces(pc));

        /* Some pre-RFC1521 gateways still use the "name=filename" convention,
         * but if a filename has already been set in the content-disposition,
         * let that take precedence, and don't set it here */
        pc = parameter_getval(ct->parameter, "name");
        if (pc && !ct->filename)
            ct->filename = m_strdup(pc);
    }

    /* Now get the subtype */
    if ((subtype = strchr (s, '/'))) {
        *subtype++ = '\0';
        for (pc = subtype; *pc && !ISSPACE(*pc) && *pc != ';'; pc++);
        ct->subtype = p_dupstr(subtype, pc - subtype);
    }

    /* Finally, get the major type */
    ct->type = mutt_check_mime_type(s);

    if (ct->type == TYPEOTHER) {
        ct->xtype = m_strdup(s);
    }

    if (!ct->subtype) {
        /* Some older non-MIME mailers (i.e., mailtool, elm) have a content-type
         * field, so we can attempt to convert the type to BODY here.
         */
        switch (ct->type) {
            char buffer[STRING];

          case TYPETEXT:
            ct->subtype = m_strdup("plain");
            break;

          case TYPEAUDIO:
            ct->subtype = m_strdup("basic");
            break;

          case TYPEMESSAGE:
            ct->subtype = m_strdup("rfc822");
            break;

          case TYPEOTHER:
            ct->type = TYPEAPPLICATION;
            snprintf(buffer, sizeof(buffer), "x-%s", s);
            ct->subtype = m_strdup(buffer);
            break;

          default:
            ct->subtype = m_strdup("x-unknown");
            break;
        }
    }

    /* Default character set for text types. */
    if (ct->type == TYPETEXT) {
        pc = parameter_getval(ct->parameter, "charset");
        if (!pc) {
            parameter_setval(&ct->parameter, "charset",
                             charset_getfirst(AssumedCharset));
        }
    }
}

static void parse_content_disposition(const char *s, BODY *ct)
{
    if (!ascii_strncasecmp(s, "inline", 6)) {
        ct->disposition = DISPINLINE;
    } else if (!ascii_strncasecmp(s, "form-data", 9)) {
        ct->disposition = DISPFORMDATA;
    } else {
        ct->disposition = DISPATTACH;
    }

    /* Check to see if a default filename was given */
    if ((s = strchr (s, ';'))) {
        parameter_t *parms = parse_parameters(vskipspaces(s));

        if ((s = parameter_getval(parms, "filename")))
            m_strreplace(&ct->filename, s);
        if ((s = parameter_getval(parms, "name")))
            ct->form_name = m_strdup(s);

        parameter_list_wipe(&parms);
    }
}

/* args:
 *	fp	stream to read from
 *
 *	digest	1 if reading subparts of a multipart/digest, 0
 *		otherwise
 */
BODY *mutt_read_mime_header(FILE *fp, int digest)
{
    BODY *body = body_new();
    char *line = p_new(char, LONG_STRING);
    ssize_t linelen = LONG_STRING;
    char *p;

    body->hdr_offset  = ftello(fp);
    body->encoding    = ENC7BIT;    /* default from RFC1521 */
    body->disposition = DISPINLINE;
    body->type        = digest ? TYPEMESSAGE : TYPETEXT;

    while (mutt_read_rfc822_line(fp, &line, &linelen)) {
        /* Find the value of the current header */
        if ((p = strchr(line, ':'))) {
            *p++ = '\0';
            p = vskipspaces(p);
            if (!*p)
                continue;
        } else {
            break;
        }

        switch (mime_which_token(line, -1)) {
          case MIME_CONTENT_TYPE:
            mutt_parse_content_type (p, body);
            break;

          case MIME_CONTENT_TRANSFER_ENCODING:
            body->encoding = mutt_check_encoding (p);
            break;

          case MIME_CONTENT_DISPOSITION:
            parse_content_disposition(p, body);
            break;

          case MIME_CONTENT_DESCRIPTION:
            m_strreplace(&body->description, p);
            rfc2047_decode(&body->description);
            break;

          default: break;
        }
    }

    body->offset = ftello(fp);       /* Mark the start of the real data */
    if (!body->subtype) {
        if (body->type == TYPETEXT)
            body->subtype = m_strdup("plain");
        if (body->type == TYPEMESSAGE)
            body->subtype = m_strdup("rfc822");
    }

    p_delete(&line);
    return (body);
}

void mutt_parse_part(FILE *fp, BODY *b)
{
    char *bound = 0;

    switch (b->type) {
      case TYPEMULTIPART:
        bound = parameter_getval(b->parameter, "boundary");
        fseeko(fp, b->offset, SEEK_SET);
        b->parts = mutt_parse_multipart(fp, bound, b->offset + b->length,
                                        mime_which_token(b->subtype, -1) == MIME_DIGEST);
        break;

      case TYPEMESSAGE:
        if (b->subtype) {
            fseeko(fp, b->offset, SEEK_SET);

            if (mutt_is_message_type(b)) {
                b->parts = mutt_parse_messageRFC822(fp, b);
            } else
            if (mime_which_token(b->subtype, -1) == MIME_EXTERNAL_BODY) {
                b->parts = mutt_read_mime_header(fp, 0);
            } else {
                return;
            }
        }
        break;

      default:
        return;
    }

    /* try to recover from parsing error */
    if (!b->parts) {
        b->type = TYPETEXT;
        m_strreplace(&b->subtype, "plain");
    }
}

/* parse a MESSAGE/RFC822 body
 *
 * args:
 *	fp		stream to read from
 *
 *	parent		structure which contains info about the message/rfc822
 *			body part
 *
 * NOTE: this assumes that `parent->length' has been set!
 */
BODY *mutt_parse_messageRFC822(FILE * fp, BODY * parent)
{
    BODY *msg;

    parent->hdr = header_new();
    parent->hdr->offset = ftello(fp);
    parent->hdr->env    = mutt_read_rfc822_header(fp, parent->hdr, 0, 0);

    msg = parent->hdr->content;

    /* ignore the length given in the content-length since it could be wrong
       and we already have the info to calculate the correct length */
    /* if (msg->length == -1) */
    /* if body of this message is empty, we can end up with a negative length */
    msg->length = MAX(0, parent->length - (msg->offset - parent->offset));

    mutt_parse_part(fp, msg);

    return msg;
}

/* parse a multipart structure
 *
 * args:
 *	fp		stream to read from
 *
 *	bound   	body separator
 *
 *	end_off		length of the multipart body (used when the final
 *			boundary is missing to avoid reading too far)
 *
 *	digest		1 if reading a multipart/digest, 0 otherwise
 */

BODY *
mutt_parse_multipart(FILE *fp, const char *bound, off_t end_off, int digest)
{
    char buffer[LONG_STRING];
    BODY *head = NULL;
    BODY **last = &head;
    int blen = m_strlen(bound);
    int final = 0;                /* did we see the ending boundary? */

    if (!blen) {
        mutt_error _("multipart message has no boundary parameter!");
        return NULL;
    }

    while (ftello(fp) < end_off && fgets(buffer, sizeof(buffer), fp)) {
        int len, crlf, i;

        len  = m_strlen(buffer);
        crlf = len > 1 && buffer[len - 2] == '\r';

        if (buffer[0] == '-' && buffer[1] == '-'
        && !m_strncmp(buffer + 2, bound, blen))
        {
            if (*last) {
                BODY *b = *last;

                /* if the body is empty, we can end up with a -1 length */
                b->length = MAX(0, ftello(fp) - b->offset - len - 1 - crlf);
                if (b->parts && b->parts->length == 0) {
                    b->parts->length = ftello(fp) - b->parts->offset
                                     - len - 1 - crlf;
                }
            }

            /* Remove any trailing whitespace, up to the length of the boundary */
            for (i = len - 1; ISSPACE(buffer[i]) && i >= blen + 2; i--)
                buffer[i] = '\0';

            /* Check for the end boundary */
            final = buffer[blen + 3] == '-' && buffer[blen + 4] == '-';
            if (final)
                break;

            if (buffer[2 + blen] == '\0') {
                BODY *new = mutt_read_mime_header(fp, digest);

                /*
                 * Consistency checking - catch
                 * bad attachment end boundaries
                 */

                if (new->offset > end_off) {
                    body_list_wipe(&new);
                    break;
                }

                if (*last)
                    last = &(*last)->next;
                *last = new;
            }
        }
    }

    /* in case of missing end boundary, set the length to something reasonable */
    if (*last && (*last)->length == 0 && !final)
        (*last)->length = end_off - (*last)->offset;

    /* parse recursive MIME parts */
    {
        BODY *b;
        for (b = head; b; b = b->next)
            mutt_parse_part(fp, b);
    }

    return (head);
}

static const char *
uncomment_timezone(char *buf, size_t buflen, const char *tz)
{
    char *p;

    if (*tz != '(')
        return tz;                  /* no need to do anything */

    tz = vskipspaces(tz + 1);
    p = strpbrk(tz, " )");
    if (!p)
        return tz;

    m_strncpy(buf, buflen, tz, p - tz);
    return buf;
}

/* parses a date string in RFC822 format:
 *
 * Date: [ weekday , ] day-of-month month year hour:minute:second timezone
 *
 * This routine assumes that `h' has been initialized to 0.  the `timezone'
 * field is optional, defaulting to +0000 if missing.
 */
time_t mutt_parse_date(const char *s, HEADER *h)
{
    int zhours = 0, zminutes = 0, zoccident = 0;
    char scratch[STRING];
    struct tm tm;
    int count = 0;
    char *p;

    /* Don't modify our argument. Fixed-size buffer is ok here since
       the date format imposes a natural limit.  */

    m_strcpy(scratch, sizeof(scratch), s);

    /* kill the day of the week, if it exists. */
    p = strchr(scratch, ',');
    p = vskipspaces(p ? p + 1 : scratch);

    p_clear(&tm, 1);

    while ((p = strtok (p, " \t")) != NULL) {
        char tzstr[STRING];
        const char *ptz;

        switch (count) {
          case 0:                    /* day of the month */
            if (!isdigit((unsigned char)*p))
                return -1;
            tm.tm_mday = atoi(p);
            if (tm.tm_mday > 31)
                return -1;
            break;

          case 1:                    /* month of the year */
            tm.tm_mon = mutt_check_month(p);
            if (tm.tm_mon < 0)
                return -1;
            break;

          case 2:                    /* year */
            tm.tm_year = atoi(p);
            if (tm.tm_year < 50)
                tm.tm_year += 100;
            else if (tm.tm_year >= 1900)
                tm.tm_year -= 1900;
            break;

          case 3:                    /* time of day */
            tm.tm_hour = strtol(p, &p, 10);
            if (*p++ != ':')
                return -1;
            tm.tm_min  = strtol(p, &p, 10);
            if (*p++ == ':') {
                tm.tm_sec = strtol(p, &p, 10);
            } else {
                tm.tm_sec = 0;
            }
            break;

          case 4:                    /* timezone */
            /* sometimes we see things like (MST) or (-0700) so attempt to
             * compensate by uncommenting the string if non-RFC822 compliant
             */
            ptz = uncomment_timezone(tzstr, sizeof(tzstr), p);

            if (*ptz == '+' || *ptz == '-') {
                if (isdigit((unsigned char)ptz[1])
                &&  isdigit((unsigned char)ptz[2])
                &&  isdigit((unsigned char)ptz[3])
                &&  isdigit((unsigned char)ptz[4]))
                {
                    zoccident = ptz[0] == '-';
                    zhours    = (ptz[1] - '0') * 10 + (ptz[2] - '0');
                    zminutes  = (ptz[3] - '0') * 10 + (ptz[4] - '0');
                }
            }
            break;
        }
        count++;
        p = NULL;
    }

    if (count < 4) {  /* don't check for missing timezone */
        return -1;
    }

    if (h) {
        h->zhours    = zhours;
        h->zminutes  = zminutes;
        h->zoccident = zoccident;
    }

    return mutt_mktime(&tm, 0) + (zoccident ? 1 : -1) * (zhours * 3600 + zminutes * 60);
}

string_list_t **mutt_parse_rfc822_line(ENVELOPE *e, HEADER *hdr, char *line, char *p,
                              short weed, short do_2047, string_list_t **user_hdrs)
{
    switch (mime_which_token(line, -1)) {
      case MIME_APPARENTLY_FROM:
        e->from = rfc822_parse_adrlist (e->from, p);
        break;

      case MIME_APPARENTLY_TO:
        e->to = rfc822_parse_adrlist (e->to, p);
        break;

      case MIME_BCC:
        e->bcc = rfc822_parse_adrlist (e->bcc, p);
        break;

      case MIME_CC:
        e->cc = rfc822_parse_adrlist (e->cc, p);
        break;

      case MIME_CONTENT_DESCRIPTION:
        if (hdr) {
            m_strreplace(&hdr->content->description, p);
            rfc2047_decode(&hdr->content->description);
        }
        break;

      case MIME_CONTENT_DISPOSITION:
        if (hdr)
            parse_content_disposition(p, hdr->content);
        break;

      case MIME_CONTENT_LENGTH:
        if (hdr) {
            if ((hdr->content->length = atoi(p)) < 0)
                hdr->content->length = -1;
        }
        break;

      case MIME_CONTENT_TRANSFER_ENCODING:
        if (hdr)
            hdr->content->encoding = mutt_check_encoding(p);
        break;

      case MIME_CONTENT_TYPE:
        if (hdr)
            mutt_parse_content_type (p, hdr->content);
        break;

      case MIME_DATE:
        m_strreplace(&e->date, p);
        if (hdr)
            hdr->date_sent = mutt_parse_date (p, hdr);
        break;

      case MIME_EXPIRES:
        if (hdr && mutt_parse_date (p, NULL) < time (NULL))
            hdr->expired = 1;
        break;

#ifdef USE_NNTP
      case MIME_FOLLOWUP_TO:
        if (!e->followup_to) {
            m_strrtrim(p);
            e->followup_to = m_strdup(skipspaces(p));
        }
        break;
#endif

      case MIME_FROM:
        e->from = rfc822_parse_adrlist(e->from, p);
        /* don't leave from info NULL if there's an invalid address (or
         * whatever) in From: field; mutt would just display it as empty
         * and mark mail/(esp.) news article as your own. aaargh! this
         * bothered me for _years_ */
        if (!e->from) {
            e->from = address_new();
            e->from->personal = m_strdup(p);
        }
        break;

      case MIME_IN_REPLY_TO:
        string_list_wipe(&e->in_reply_to);
        e->in_reply_to = mutt_parse_references(p, 1);
        break;

      case MIME_LINES:
        if (hdr) {
            /* HACK - mutt has, for a very short time, produced negative
               Lines header values.  Ignore them. */
            hdr->lines = MAX(0, atoi(p));
        }
        break;

      case MIME_LIST_POST:
        /* RFC 2369.  FIXME: We should ignore whitespace, but don't. */
        if (m_strncmp(p, "NO", 2)) {
            char *beg, *end;

            for (beg = strchr (p, '<'); beg; beg = strchr (end, ',')) {
                ++beg;
                if (!(end = strchr (beg, '>')))
                    break;

                /* Take the first mailto URL */
                if (url_check_scheme (beg) == U_MAILTO) {
                    p_delete(&e->list_post);
                    e->list_post = p_dupstr(beg, end - beg);
                    break;
                }
            }
        }
        break;

      case MIME_MAIL_FOLLOWUP_TO:
        e->mail_followup_to = rfc822_parse_adrlist(e->mail_followup_to, p);
        break;

      case MIME_MAIL_REPLY_TO:
        address_list_wipe(&e->reply_to);
        e->reply_to = rfc822_parse_adrlist(e->reply_to, p);
        break;

      case MIME_MESSAGE_ID:
        {
            const char *beg, *end;

            /* We add a new "Message-ID:" when building a message */
            p_delete(&e->message_id);

            if ((beg = strchr(p, '<')) && (end = strchr(beg, '>')))
                e->message_id = p_dupstr(beg, (end - beg) + 1);
        }
        break;

      case MIME_MIME_VERSION:
        if (hdr)
            hdr->mime = 1;
        break;

#ifdef USE_NNTP
      case MIME_NEWSGROUPS:
        p_delete(&e->newsgroups);
        m_strrtrim(p);
        e->newsgroups = m_strdup(skipspaces(p));
        break;
#endif

      case MIME_ORGANIZATION:
        if (!e->organization && mime_which_token(p, -1) == MIME_UNKNOWN)
            e->organization = m_strdup(p);
        break;

      case MIME_RECEIVED:
        if (hdr && !hdr->received) {
            char *d = strchr(p, ';');
            if (d)
                hdr->received = mutt_parse_date(d + 1, NULL);
        }
        break;

      case MIME_REFERENCES:
        string_list_wipe(&e->references);
        e->references = mutt_parse_references(p, 0);
        break;

      case MIME_REPLY_TO:
        e->reply_to = rfc822_parse_adrlist(e->reply_to, p);
        break;

      case MIME_RETURN_PATH:
        e->return_path = rfc822_parse_adrlist(e->return_path, p);
        break;

      case MIME_SENDER:
        e->sender = rfc822_parse_adrlist (e->sender, p);
        break;

      case MIME_STATUS:
        if (hdr) {
            while (*p) {
                switch (*p) {
                  case 'r':
                    hdr->replied = 1;
                    break;
                  case 'O':
                    hdr->old = 1;
                    break;
                  case 'R':
                    hdr->read = 1;
                    break;
                }
                p++;
            }
        }
        break;

      case MIME_SUBJECT:
        if (!e->subject)
            e->subject = m_strdup(p);
        break;

      case MIME_SUPERCEDES:
      case MIME_SUPERSEDES:
        if (hdr)
            e->supersedes = m_strdup(p);
        break;

      case MIME_TO:
        e->to = rfc822_parse_adrlist(e->to, p);
        break;

#ifdef USE_NNTP
      case MIME_X_COMMENT_TO:
        if (!e->x_comment_to)
            e->x_comment_to = m_strdup(p);
        break;
#endif

      case MIME_X_LABEL:
        e->x_label = m_strdup(p);
        break;

#ifdef USE_NNTP
      case MIME_XREF:
        if (!e->xref)
            e->xref = m_strdup(p);
        break;
#endif

      case MIME_X_STATUS:
        if (hdr) {
            while (*p) {
                switch (*p) {
                  case 'A':
                    hdr->replied = 1;
                    break;
                  case 'D':
                    hdr->deleted = 1;
                    break;
                  case 'F':
                    hdr->flagged = 1;
                    break;
                  default:
                    break;
                }
                p++;
            }
        }
        break;

      default:
        if (!user_hdrs)
            break;

        /* restore the original line */
        line[m_strlen(line)] = ':';

        if (weed && string_list_contains(Ignore, line, "*")
        && !string_list_contains(UnIgnore, line, "*")) {
            break;
        }

        *user_hdrs = string_item_new();
        (*user_hdrs)->data = m_strdup(line);
        if (do_2047)
            rfc2047_decode(&(*user_hdrs)->data);
        return &(*user_hdrs)->next;
    }

    return user_hdrs;
}

/* mutt_read_rfc822_header() -- parses a RFC822 header
 *
 * Args:
 *
 * f		stream to read from
 *
 * hdr		header structure of current message (optional).
 *
 * user_hdrs	If set, store user headers.  Used for recall-message and
 * 		postpone modes.
 *
 * weed		If this parameter is set and the user has activated the
 * 		$weed option, honor the header weed list for user headers.
 * 	        Used for recall-message.
 *
 * Returns:     newly allocated envelope structure.  You should free it by
 *              envelope_delete() when envelope stay unneeded.
 */
ENVELOPE *
mutt_read_rfc822_header(FILE *f, HEADER *hdr, short user_hdrs, short weed)
{
    ENVELOPE *e = envelope_new();
    string_list_t **last = user_hdrs ? &e->userhdrs : NULL;

    char *line = p_new(char, LONG_STRING);
    ssize_t linelen = LONG_STRING;
    off_t loc;

    if (hdr && !hdr->content) {
        hdr->content = body_new();

        /* set the defaults from RFC1521 */
        hdr->content->type     = TYPETEXT;
        hdr->content->subtype  = m_strdup("plain");
        hdr->content->encoding = ENC7BIT;
        hdr->content->length   = -1;

        /* RFC 2183 says this is arbitrary */
        hdr->content->disposition = DISPINLINE;
    }

    while ((loc = ftello(f)),
           mutt_read_rfc822_line(f, &line, &linelen))
    {
        char buf[LONG_STRING + 1] = "";
        char *p;

        p = strpbrk(line, ": \t");
        if (!p || *p != ':') {
            /* some bogus MTAs will quote the original "From " line */
            if (!m_strncmp(">From ", line, 6) || !m_strncmp("From ", line, 5))
                continue;               /* just ignore */

            fseeko(f, loc, 0);
            break;                    /* end of header */
        }

        if (rx_list_match2(SpamList, line, buf, sizeof(buf))
        && !rx_list_match(NoSpamList, line))
        {
            /* if spam tag already exists, figure out how to amend it */
            if (e->spam && *buf) {
                if (SpamSep) {
                    /* If SpamSep defined, append with separator */
                    mutt_buffer_addstr(e->spam, SpamSep);
                    mutt_buffer_addstr(e->spam, buf);
                } else {
                    /* else overwrite */
                    mutt_buffer_reset(e->spam);
                    mutt_buffer_addstr(e->spam, buf);
                }
            }

            if (!e->spam) {
                e->spam = mutt_buffer_from(NULL, buf);
            }
        }

        *p++ = '\0';
        p = vskipspaces(p);
        if (!*p)
            continue;                 /* skip empty header fields */

        last = mutt_parse_rfc822_line(e, hdr, line, p, weed, 1, last);
    }

    p_delete(&line);

    if (hdr) {
        hdr->content->hdr_offset = hdr->offset;
        hdr->content->offset     = ftello(f);
        rfc2047_decode_envelope(e);
        /* check for missing or invalid date */
        if (hdr->date_sent <= 0) {
            hdr->date_sent = hdr->received;
        }
    }

    return e;
}

/* Compares mime types to the ok and except lists */
static int count_body_parts_check(string_list_t **checklist, BODY *b)
{
    string_list_t *type;

    for (type = *checklist; type; type = type->next) {
        ATTACH_MATCH *a = (ATTACH_MATCH *)type->data;

        if ((a->major_int == TYPEANY || a->major_int == b->type)
        &&  !regexec(&a->minor_rx, b->subtype, 0, NULL, 0)) {
            return 1;
        }
    }

    return 0;
}

static int count_body_parts (BODY *body, int flags)
{
    int count = 0;
    BODY *bp;

    if (!body)
        return 0;

    for (bp = body; bp != NULL; bp = bp->next) {
        /* Initial disposition is to count and not to recurse this part. */
        int shallcount, shallrecurse, iscontainer;
        int tok = mime_which_token(bp->subtype, -1);

        iscontainer  = bp->type == TYPEMESSAGE || bp->type == TYPEMULTIPART;

        /* don't recurse in external bodies or multipart/alternatives */
        shallrecurse = (bp->type == TYPEMESSAGE && tok != MIME_EXTERNAL_BODY)
                    || (bp->type == TYPEMULTIPART && tok != MIME_ALTERNATIVE);

        /* Don't count top level containers and fundamental inlines */
        shallcount   = !(iscontainer && (flags & M_PARTS_TOPLEVEL))
                    && !(!iscontainer && bp->disposition == DISPINLINE && bp == body);

        if (shallcount) {
            /* Turn off shallcount if message type is not in ok list,
             * or if it is in except list. Check is done separately for
             * inlines vs. attachments.
             */

            if (bp->disposition == DISPATTACH) {
                if (!count_body_parts_check(&AttachAllow, bp))
                    shallcount = 0;
                if (count_body_parts_check(&AttachExclude, bp))
                    shallcount = 0;
            } else {
                if (!count_body_parts_check(&InlineAllow, bp))
                    shallcount = 0;
                if (count_body_parts_check(&InlineExclude, bp))
                    shallcount = 0;
            }
        }

        bp->attach_qualifies = shallcount;
        count += shallcount;

        if (shallrecurse) {
            bp->attach_count = count_body_parts(bp->parts,
                                                flags & ~M_PARTS_TOPLEVEL);
            count += bp->attach_count;
        }
    }

    return count;
}

int mutt_count_body_parts(HEADER *hdr, int flags)
{
    if (hdr->attach_valid && !(flags & M_PARTS_RECOUNT))
        return hdr->attach_total;

    if (AttachAllow || AttachExclude || InlineAllow || InlineExclude)
        hdr->attach_total = count_body_parts(hdr->content,
                                             flags | M_PARTS_TOPLEVEL);
    else
        hdr->attach_total = 0;

    hdr->attach_valid = 1;
    return hdr->attach_total;
}
